local class_name = "standalone_damage_popups"
local loaded = rawget(_G, class_name)
local DMGPopup = loaded or rawset(_G, class_name, {
    DamagePop = class(),
    damage_pops = {},
    stationary_pop = {},
    finished_pops = {},
    _created_t = {},
    unit_position = {},
    show_all_player_pops = true,
    show_ai_popup = false,
    use_player_name = true,
    stationary_damage_pop = true,
    death_sign = "",
    font = "fonts/font_medium_mf",
    font_size = 35,
    pop_margin = 200,
    lifetime = 0.7,
    crit_color = "purple",
    headshot_color = "warm_yellow",
    friendly_color = "light_blue",
    civilian_color = "cobalt_blue",
    enemy_color = "leaf_green",
    special_enemy_color = "leaf_green",
    death_color = "blood_red",
    immortal_color = "gray_blue",
    sentry_gun_color = "warm_yellow",
    colors = {
        blood_red = "db1414",
        orange = "faa200",
        purple = "8900eb",
        pink = "eb00d7",
        gray_blue = "3d7bae",
        light_blue = "05d1ff",
        bone_white = "eee5c4",
        leaf_green = "0db531",
        warm_yellow = "e6d200",
        cobalt_blue = "0081eb",
        nothing = "000000"
    },
    peer_colors = {
		"c2fc97", --1
		"78b7cc", --2
		"b26859", --3
		"cca166", --4
		"33ccff" --ai
	}
}) and _G[class_name]
 
if not loaded then
    function DMGPopup:init()
        math.randomseed(os.time())

        if WolfHUD then
            WolfHUD.settings.DamagePopup.DISPLAY_MODE = 1 -- disable wolfhud dmg popup
        end

        if VHUDPlus then
            VHUDPlus.settings.DamagePopupNew.DISPLAY_MODE = 1 -- disable VHUDPlus dmg popup
        end
    end

    DMGPopup:init()

    function DMGPopup:add_damage_pop(unit, attack_data, is_husk, peer_data)
        if not managers.hud or game_state_machine and not game_state_machine:verify_game_state(GameStateFilters.any_ingame_playing) then
            return
        end

        local unit_damage = unit:character_damage()
        local col_ray = attack_data.col_ray or {}
        local damage_ceil, damage_decimal = math.modf((attack_data.damage or attack_data.raw_damage or 0))
        local ceil_mod, decimal_mod = damage_ceil * 10, damage_decimal * 10
        local damage = tostring((ceil_mod < 1 and math.ceil(decimal_mod) or ceil_mod))
        local format, stationary_pop = {}, false

        format[#format + 1] = is_husk and {{self.colors[self.enemy_color], self.peer_colors[peer_data.peer_id]}, {damage, " - " .. peer_data.peer_name}} or nil
        format[#format + 1] = (unit_damage._immortal or unit_damage._invulnerable) and {{self.colors[self.immortal_color]}, {damage .. " - IMMORTAL"}} or nil
        format[#format + 1] = unit:in_slot(16) and {{self.colors[self.friendly_color], self.colors[self.death_color]}, {damage, "!"}} or nil
        format[#format + 1] = unit.base and CopDamage.is_civilian(unit:base()._tweak_table) and {{self.colors[self.civilian_color]}, {damage .. " - Oof!"}} or nil
        format[#format + 1] = (attack_data.type == "death" or unit_damage._dead) and {{self.colors[self.headshot_color], self.colors[self.death_color], self.colors[self.headshot_color]}, {self.death_sign, damage, self.death_sign}} or nil
        format[#format + 1] = attack_data.critical_hit and {{self.colors[self.crit_color], self.colors[self.headshot_color]}, {damage, self.death_sign}} or nil
        format[#format + 1] = unit_damage.is_head and unit_damage:is_head(col_ray.body) and {{self.colors[self.headshot_color], self.colors[self.death_color]}, {damage, "!"}} or nil
        format[#format + 1] = managers.groupai:state():is_enemy_special(unit) and {{self.colors[self.special_enemy_color]}, {damage}} or nil
        format[#format + 1] = unit:in_slot(25) and {{self.colors[self.sentry_gun_color]}, {damage}} or nil
        format[#format + 1] = {{self.colors[self.enemy_color]}, {damage}}

        local colors, texts = unpack(format[1])
        local formated_text = ""
        self.color_data = {}
        
        for k, text in ipairs(texts) do
            local from = utf8.len(formated_text)
            formated_text = formated_text .. text
            table.insert(self.color_data, {from = from, to = utf8.len(formated_text), color = Color(colors[k])})
        end

        if formated_text:match("%d+") and not formated_text:match("[^%d+]") and self.stationary_damage_pop then
            stationary_pop = true
        end

        self.unit_id = stationary_pop and class_name or Idstring(math.random(1, 100000))
        local id = stationary_pop and unit_id or Idstring(math.random(1, 100000))
        self.unit_position[self.unit_id] = col_ray.position or attack_data.pos or col_ray.hit_position or unit:position()
        self.damage_pops[self.unit_id] = self.damage_pops[self.unit_id] or {}

        if not self.damage_pops[self.unit_id][id] then
            self.damage_pops[self.unit_id][id] = DMGPopup.DamagePop:new(formated_text, self.color_data[1].color, self.unit_id, stationary_pop)
        end
    end

    function DMGPopup:is_valid_pop(attack_data)
        local attacker_unit = attack_data and attack_data.attacker_unit
        local attacker_base = attacker_unit and attacker_unit:base()
        local is_ai = attacker_unit and attacker_unit:in_slot(16)

        if attacker_base and (not is_ai or self.show_ai_popup) then
            if attacker_base.thrower_unit then
                attacker_unit = attacker_base:thrower_unit()
            elseif attacker_base.sentry_gun then
                attacker_unit = attacker_base:get_owner()
            end
            
            if attacker_unit then
                local session = managers.network:session()
                local peer = session and session:peer_by_unit(attacker_unit)
                local is_players = attacker_unit:in_slot(3, 5)
            
                return {peer_name = (is_ai and tweak_data.peer_colors[5]:sub(3):upper() or peer and (not self.use_player_name and tweak_data.peer_colors[peer:id()]:sub(3):upper() or peer:name():upper()) or tweak_data.peer_colors[1]:sub(3):upper()), peer_id = (is_ai and 5 or peer and peer:id() or 1)}, ((is_ai or is_players) and attacker_unit ~= managers.player:player_unit() or false), attacker_unit
            end
        end
        return {}
    end

    function DMGPopup:update(t, dt)
        local cam = managers.viewport:get_current_camera()
        self._t = t

        for unit_id in pairs(cam and self.damage_pops or {}) do
            for _, pop in pairs(self.damage_pops[unit_id] or {}) do
                if (pop.dead and pop.dead[unit_id]) then
                    self._panel:remove(pop._panel[unit_id])
                elseif pop.update then
                    pop:update(t, cam)
                end
            end
        end
    end

    function DMGPopup.DamagePop:init(formated_text, color_data, unit_id, stationary_pop)
        DMGPopup._hud = DMGPopup._hud or managers.hud:script(PlayerBase.PLAYER_INFO_HUD_FULLSCREEN_PD2)
        DMGPopup._ws = DMGPopup._ws or managers.hud._workspace
        DMGPopup._panel = DMGPopup._panel or DMGPopup._hud and DMGPopup._hud.panel or DMGPopup._ws:panel({name = class_name})
        self.unit_id = unit_id
        DMGPopup._created_t[self.unit_id] = DMGPopup._t
        self.font_size = DMGPopup.font_size
        self.font = DMGPopup.font
        self.formated_text = formated_text
        self.stationary_pop = stationary_pop
        self.dead = {}
        self._panel = {}
        self._panel[self.unit_id] = DMGPopup._panel:panel({
            name = "panel" .. self.unit_id,
            visible = false
        })
        local text = self._panel[self.unit_id]:text({
            name = "text",
            text = self.formated_text,
            font = self.font,
            font_size = self.font_size,
            color = color_data,
            blend_mode = "add"
        })
        local x, y, w, h = text:text_rect()

        self._panel[self.unit_id]:set_size(w, h)
        text:set_alpha(1)
    end
    
    function DMGPopup.DamagePop:update(t, cam)
        local life_left = (t - (DMGPopup._created_t[self.unit_id] or t)) / DMGPopup.lifetime + (self.stationary_damage_pop and -0.2 or 0)
        
        if life_left > 1 then
            self.dead[self.unit_id] = true
        else
            local panel = self._panel[self.unit_id]
            local child = panel:child("text")
            local cam_forward = Vector3()
            local world_pos = Vector3()
            local screen_pos = Vector3()

            mrotation.y(cam:rotation(), cam_forward)
            mvector3.set(world_pos, DMGPopup.unit_position[self.unit_id])
            mvector3.set(screen_pos, DMGPopup._ws:world_to_screen(cam, world_pos))
            mvector3.subtract(world_pos, cam:position())
            mvector3.normalize(world_pos)
            panel:set_center(screen_pos.x, screen_pos.y - 2 * panel:h())
            panel:set_visible(mvector3.dot(cam_forward, world_pos) >= 0)
            panel:set_alpha(-life_left+1.1)
            
            local x,y,w,h = child:text_rect()

            if self.stationary_pop then
                child:set_text(self.formated_text)
            else
                panel:set_y(-(life_left*DMGPopup.pop_margin)+y)
            end

            for k, data in pairs(DMGPopup.color_data) do
                if DMGPopup.finished_pops[self.unit_id] then
                    break
                end

                child:set_range_color(data.from, data.to, data.color)

                if #DMGPopup.color_data == k then
                    DMGPopup.finished_pops[self.unit_id] = true
                end
            end
        end
    end
end

if RequiredScript:lower() == "lib/managers/playermanager" then
    local orig_update = PlayerManager.update
    function PlayerManager:update(...)
        orig_update(self, ...)

        if not managers.hud or game_state_machine and not game_state_machine:verify_game_state(GameStateFilters.any_ingame_playing) then
            return
        end

        DMGPopup:update(...)
    end

    local orig_on_damage_dealt = PlayerManager.on_damage_dealt
    function PlayerManager:on_damage_dealt(unit, attack_data, ...)
        local peer_data, is_husk, attacker_unit = DMGPopup:is_valid_pop(attack_data)
        
        if peer_data.peer_name and peer_data.peer_id and (not DMGPopup.show_all_player_pops or unit:in_slot(25)) then
            DMGPopup:add_damage_pop(unit, attack_data, is_husk, peer_data)
        end

        orig_on_damage_dealt(self, unit, attack_data, ...)
    end
elseif RequiredScript:lower() == "lib/units/enemies/cop/copdamage" then
    local orig_cop_on_damage_received = CopDamage._on_damage_received
    function CopDamage:_on_damage_received(attack_data, ...)
        local peer_data, is_husk, attacker_unit = DMGPopup:is_valid_pop(attack_data)

        if peer_data.peer_name and peer_data.peer_id and DMGPopup.show_all_player_pops then
            DMGPopup:add_damage_pop(self._unit, attack_data, is_husk, peer_data)
        end

        orig_cop_on_damage_received(self, attack_data, ...)
    end
elseif RequiredScript:lower() == "lib/units/civilians/civiliandamage" then
    local orig_civ_on_damage_received = CivilianDamage._on_damage_received
    function CivilianDamage:_on_damage_received(attack_data, ...)
        local peer_data, is_husk, attacker_unit = DMGPopup:is_valid_pop(attack_data)

        if peer_data.peer_name and peer_data.peer_id and DMGPopup.show_all_player_pops then
            DMGPopup:add_damage_pop(self._unit, attack_data, is_husk, peer_data)
        end

        orig_civ_on_damage_received(self, attack_data, ...)
    end
end